import sys
import os
import copy
import random
import networkx as nx
import numpy as np

# lib_path = './lib'
# sys.path.insert(0, lib_path)
import MultiContagion as mc
import helpfunctions as  hf


def make_random_network_layers_and_coupling(n_countries, n_layers, mean_degree, \
    network_type="regular", coupling="simple"):
    """
    Function that makes the list of adjacency matrix for layer and the coupling
    between them. Layers can be Erdos-Renyi or Regular.
    Args:
        n_countries(int)
        n_layer(int)
        mean_degree(int): for now all layers have same degree
        network_type(str): defines type of null model for layers. So far
        accepted values are either regular or erdos
        coupling(str): defines type coupling which gives vertical contagion rule
        So far accepted values are "simple" which is unweighted or
        "degree_weighted" which takes into account the indegree of each layer
    Returns:
        A_list(list of np array): list of adjacency matrices
        C(np array): coupling array
    """
    assert(network_type=="regular" or network_type=="erdos")
    assert(coupling=="simple" or coupling=="degree_weighted")
    assert(mean_degree<n_countries)

    A_list = []
    p = mean_degree / n_countries
    for i in range(n_layers):
        if network_type == "regular":
            assert(n_countries*mean_degree%2 == 0)
            g = nx.random_regular_graph(mean_degree, n_countries)
        elif network_type == "erdos":
            g = nx.erdos_renyi_graph(n_countries, p)
        a = nx.to_numpy_array(g)
        A_list.append(a)

    if coupling == "simple":
        C = np.ones([n_countries, n_layers])
    elif coupling == "degree_weighted":
        C = np.zeros([n_countries, n_layers])
        #C_outdegree = np.zeros([n_countries, n_layers])
        for l in range(n_layers):
            C[:, l] = np.sum(A_list[l], axis=0)
            #C_outdegree[:, l] = np.sum(A_list[l], axis=1)

    return A_list, C



def make_modular(n_countries, n_layers, mean_degree, \
    network_type="regular", coupling="simple"):
    """
    Function that makes a modular network.
    Args:
        n_countries(int)
        n_layer(int)
        mean_degree(int)
        network_type(str): defines type of null model for layers. So far
        accepted values are either regular or erdos
        coupling(str): defines type coupling which gives vertical contagion rule
        So far accepted values are "simple" which is unweighted or
        "degree_weighted" which takes into account the indegree of each layer
    Returns:
        A_list(list of np array): list of adjacency matrices
        C(np array): coupling array

    """
    assert(network_type=="regular" or network_type=="erdos")
    assert(coupling=="simple" or coupling=="degree_weighted")
    assert(mean_degree<n_countries)

    A_list = []
    p = mean_degree / n_countries
    for i in range(n_layers):
        if network_type == "regular":
            assert(n_countries*mean_degree%2 == 0)
            g = nx.random_regular_graph(mean_degree, n_countries)
        elif network_type == "erdos":
            g = nx.erdos_renyi_graph(n_countries, p)
        a = nx.to_numpy_array(g)
        A_list.append(a)

    if coupling == "simple":
        C = np.ones([n_countries, n_layers])
    elif coupling == "degree_weighted":
        C = np.zeros([n_countries, n_layers])
        #C_outdegree = np.zeros([n_countries, n_layers])
        for l in range(n_layers):
            C[:, l] = np.sum(A_list[l], axis=0)
            #C_outdegree[:, l] = np.sum(A_list[l], axis=1)

    return A_list, C

def run_contagion_simulation(states, n_countries, n_layers, mean_degree,
    tau_hor, tau_ver, type_ver="layer", network_type="regular", coupling="simple"):
    """
    run contagion for intitial seed nodes (states)
    Args:

    Return:
        (rho):final fraction of activated nodes,
        (rho_hor):final fraction of horizontally activated nodes
        (rho_ver):final fraction of vertically activated nodes
    """
    A_list, C = make_random_network_layers_and_coupling(n_countries, n_layers, mean_degree,\
    network_type=network_type, coupling=coupling)

    # since we are using code orginally done for countries
    names= [str(i) for i in range(n_countries)]

    # run multilayer contagion
    state_history, hor_contagion, ver_contagion, mix_contagion, t_end = \
    hf.run_contagion(A_list, C, states, hf.default_horizontal_homogenous\
    , hf.default_vertical_homogenous, tau_hor, tau_ver, type_ver=type_ver, \
    save_csv=False, \
    save_name="results/csv_contagion2019/ER_mult_test" + str(tau_hor) + \
    "tauver_"+str(tau_ver), country_names=names)

    # WARNING rho_hor + rho_ver can be less than total rho since we
    # don't add rho_0 to either vertical or horizontal contagion

    rho = hf.node_layer_defaulted(state_history) / ( n_countries * n_layers)
    rho_hor = sum(hor_contagion) / ( n_countries * n_layers)
    rho_ver = sum(ver_contagion) / ( n_countries * n_layers)

    return rho, rho_hor, rho_ver


def run_contagion_simulation_fast(states, n_countries, n_layers, mean_degree,
    tau_hor, tau_ver, network_type="regular", coupling="simple"):
    """
    same as before but counts states instead of state_history, does not record
    history
    run contagion for intitial seed nodes (states)
    Args:

    Return:
        (rho):final fraction of activated nodes,
        (rho_hor):final fraction of horizontally activated nodes
        (rho_ver):final fraction of vertically activated nodes
    """
    A_list, C = make_random_network_layers_and_coupling(n_countries, n_layers, mean_degree,\
    network_type=network_type, coupling=coupling)

    # since we are using code orginally done for countries
    names= [str(i) for i in range(n_countries)]

    # run multilayer contagion
    states, hor_contagion, ver_contagion, mix_contagion, t_end = \
    hf.run_contagion_fast(A_list, C, states, hf.default_horizontal_homogenous\
    , hf.default_vertical_homogenous, tau_hor, tau_ver, save_csv=False, \
    save_name="results/csv_contagion2019/ER_mult_test" + str(tau_hor) + \
    "tauver_"+str(tau_ver), country_names=names)

    # WARNING rho_hor + rho_ver can be less than total rho since we
    # don't add rho_0 to either vertical or horizontal contagion

    rho = hf.node_layer_defaulted_fast(states) / ( n_countries * n_layers)
    rho_hor = sum(hor_contagion) / ( n_countries * n_layers)
    rho_ver = sum(ver_contagion) / ( n_countries * n_layers)

    return rho, rho_hor, rho_ver

def initial_states(n_countries, n_layers, rho_0, spread="perlayer"):
    """
    initial states for randomly chosen node-layers as seeds
    seeds can be spread per layer, meaning rho_0*n seeds per layer or
    uniformly, which means rho_0*n seeds in total spread among layers
    """
    # TODO add diff rho_0 in each layer
    if spread == "perlayer":
        n_seeds_perlay = int(n_countries*rho_0)
        states = np.zeros([n_countries, n_layers])
        for l in range(n_layers):
            node_seeds = random.sample([i for i in range(n_countries)]\
                                        , n_seeds_perlay)
            for i in node_seeds:
                states[i,l] = 1
        
    elif spread == "uniform":
        n_seeds = int(n_countries*rho_0)
        all_node_layers = []
        for i in range(n_countries):
            for l in range(n_layers):
                all_node_layers.append((i, l))
        # randomly sample n_seeds from all the node layers
        # TODO NOTE the seeds are randomly distributed among layers
        seeds = random.sample(all_node_layers, n_seeds)
        states = np.zeros([n_countries, n_layers])
        for (i,l) in seeds:
            states[i,l] = 1

    return states

def run_several_simulations(n_simulations, n_countries, n_layers, rho_0, \
    mean_degree, tau_hor, tau_ver, network_type="regular", coupling="simple"):
    """
    function that runs several simulations for fixed parameters and returns
    average of contagion size for total, horizontal and vertical
    """
    rho_mean, rho_hor_mean, rho_ver_mean = 0,0,0
    for t in range(n_simulations):
        states = initial_states(n_countries, n_layers, rho_0)

        rho, rho_hor, rho_ver = run_contagion_simulation(states, n_countries,\
            n_layers, mean_degree, tau_hor, tau_ver, network_type=network_type,\
            coupling=coupling)

        rho_mean += rho
        rho_hor_mean += rho_hor
        rho_ver_mean += rho_ver

    return rho_mean/n_simulations, rho_hor_mean/n_simulations, rho_ver_mean/n_simulations


def run_several_simulations_fast(n_simulations, n_countries, n_layers, rho_0, \
    mean_degree, tau_hor, tau_ver, network_type="regular", coupling="simple"):
    """
    function that runs several simulations for fixed parameters and returns
    average of contagion size for total, horizontal and vertical
    """
    rho_mean, rho_hor_mean, rho_ver_mean = 0,0,0
    for t in range(n_simulations):
        states = initial_states(n_countries, n_layers, rho_0)

        rho, rho_hor, rho_ver = run_contagion_simulation_fast(states, n_countries,\
            n_layers, mean_degree, tau_hor, tau_ver, network_type=network_type,\
            coupling=coupling)

        rho_mean += rho
        rho_hor_mean += rho_hor
        rho_ver_mean += rho_ver

    return rho_mean/n_simulations, rho_hor_mean/n_simulations, rho_ver_mean/n_simulations
