import copy
import numpy as np
from scipy.special import binom
from scipy.special import factorial


def p(k, z):
    return (np.exp(-z)*(z**k)) / np.math.factorial(k)

def p_reg(k, z):
    if k == z:
        return 1.
    else:
        return 0.

def F(m,k, ϕ):
    """threshold rule
    Args:
        m(int or array): infected neighbors
        k(int or array): degree
        ϕ(int): threshold
    """
    indicator = np.heaviside([m/k - ϕ], 0)[0]
    # assert(indicator.shape == )
    return indicator

def G(ρ_0, q, z, p, F, ϕ, k_max):
    """
    Args:
        ρ_0(float): fraction of infected seeds
        q(int/array): current estimated prob of activation
        x(int): mean degree
        p(function): degree distribution
        F(function): threshold function
        ϕ(int): threshold
        k_max(int): maximum possible degree
    Return:
        g(int/array)
    """
    g = 0
    for k in range(1,k_max):
        degree_term = k * p(k, z)/z
        # sum must be until k-1, remember python is non inclusive on upper limit
        for m in range(k):
            g += degree_term * binom(k-1, m) * q**m * (1 - q)**(k-1-m) * F(m,k,ϕ)
    return ρ_0 + (1 - ρ_0)*g


def compute_q(G, ρ_0, q, z, p, F, ϕ, max_k, n_iter=None, error_tolerance=None):
    """
    q_update(function):
    """
    # specify either number of iterations or error tolerance
    assert(n_iter==None or error_tolerance==None)
    assert(n_iter!=None or error_tolerance!=None)

    q_old = copy.deepcopy(ρ_0)
    q_new = G(ρ_0, q, z, p, F, ϕ, max_k)
    if error_tolerance != None:
        while abs(q_old - q_new) > error_tolerance:
            q_old = copy.deepcopy(q_new)
            q_new =  G(ρ_0, q_old, z, p, F, ϕ, max_k)


    elif n_iter != None:
        for i in range(n_iter):
            q_old = copy.deepcopy(q_new)
            q_new =  G(ρ_0, q_old, z, p, F, ϕ, max_k)
            print(q_new)

    return q_new


def compute_rho(ρ_0, q, z, k_max, F, ϕ):
    sum = 0
    for k in range(1,k_max):
        degree_term = p(k, z)
        # sum must be until k, remember python is non inclusive on upper limit
        for m in range(k+1):
            sum += degree_term * binom(k, m) * q**m * (1 - q)**(k-m) * F(m,k,ϕ)
    return ρ_0 + (1 - ρ_0) * sum
